@@ -34,24 +34,37 @@
 
 TextLayer::TextLayer() :
     SingleColourLayer(),
-    m_model(0),
     m_editing(false),
     m_originalPoint(0, 0.0, tr("Empty Label")),
     m_editingPoint(0, 0.0, tr("Empty Label")),
-    m_editingCommand(0)
+    m_editingCommand(nullptr)
 {
     
 }
 
-void
-TextLayer::setModel(TextModel *model)
+int
+TextLayer::getCompletion(LayerGeometryProvider *) const
 {
-    if (m_model == model) return;
-    m_model = model;
+    auto model = ModelById::get(m_model);
+    if (model) return model->getCompletion();
+    else return 0;
+}
 
-    connectSignals(m_model);
+void
+TextLayer::setModel(ModelId modelId)
+{
+    auto newModel = ModelById::getAs<TextModel>(modelId);
+    
+    if (!modelId.isNone() && !newModel) {
+        throw std::logic_error("Not a TextModel");
+    }
+    
+    if (m_model == modelId) return;
+    m_model = modelId;
 
-//    SVDEBUG << "TextLayer::setModel(" << model << ")" << endl;
+    if (newModel) {
+        connectSignals(m_model);
+    }
 
     emit modelReplaced();
 }
@@ -108,29 +121,30 @@ TextLayer::isLayerScrollable(const LayerGeometryProvider *v) const
     return !v->shouldIlluminateLocalFeatures(this, discard);
 }
 
-
-TextModel::PointList
+EventVector
 TextLayer::getLocalPoints(LayerGeometryProvider *v, int x, int y) const
 {
-    if (!m_model) return TextModel::PointList();
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return {};
 
-    sv_frame_t frame0 = v->getFrameForX(-150);
-    sv_frame_t frame1 = v->getFrameForX(v->getPaintWidth() + 150);
+    int overlap = ViewManager::scalePixelSize(150);
     
-    TextModel::PointList points(m_model->getPoints(frame0, frame1));
+    sv_frame_t frame0 = v->getFrameForX(-overlap);
+    sv_frame_t frame1 = v->getFrameForX(v->getPaintWidth() + overlap);
+    
+    EventVector points(model->getEventsSpanning(frame0, frame1 - frame0));
 
-    TextModel::PointList rv;
+    EventVector rv;
     QFontMetrics metrics = QFontMetrics(QFont());
 
-    for (TextModel::PointList::iterator i = points.begin();
-         i != points.end(); ++i) {
+    for (EventVector::iterator i = points.begin(); i != points.end(); ++i) {
 
-        const TextModel::Point &p(*i);
+        Event p(*i);
 
-        int px = v->getXForFrame(p.frame);
-        int py = getYForHeight(v, p.height);
+        int px = v->getXForFrame(p.getFrame());
+        int py = getYForHeight(v, p.getValue());
 
-        QString label = p.label;
+        QString label = p.getLabel();
         if (label == "") {
             label = tr("<no text>");
         }
@@ -146,7 +160,7 @@ TextLayer::getLocalPoints(LayerGeometryProvider *v, int x, int y) const
 
         if (x >= px && x < px + rect.width() &&
             y >= py && y < py + rect.height()) {
-            rv.insert(p);
+            rv.push_back(p);
         }
     }
 
@@ -154,22 +168,23 @@ TextLayer::getLocalPoints(LayerGeometryProvider *v, int x, int y) const
 }
 
 bool
-TextLayer::getPointToDrag(LayerGeometryProvider *v, int x, int y, TextModel::Point &p) const
+TextLayer::getPointToDrag(LayerGeometryProvider *v, int x, int y, Event &p) const
 {
-    if (!m_model) return false;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return false;
 
-    sv_frame_t a = v->getFrameForX(x - 120);
-    sv_frame_t b = v->getFrameForX(x + 10);
-    TextModel::PointList onPoints = m_model->getPoints(a, b);
+    sv_frame_t a = v->getFrameForX(x - ViewManager::scalePixelSize(120));
+    sv_frame_t b = v->getFrameForX(x + ViewManager::scalePixelSize(10));
+    EventVector onPoints = model->getEventsWithin(a, b);
     if (onPoints.empty()) return false;
 
     double nearestDistance = -1;
 
-    for (TextModel::PointList::const_iterator i = onPoints.begin();
+    for (EventVector::const_iterator i = onPoints.begin();
          i != onPoints.end(); ++i) {
 
-        double yd = getYForHeight(v, (*i).height) - y;
-        double xd = v->getXForFrame((*i).frame) - x;
+        double yd = getYForHeight(v, i->getValue()) - y;
+        double xd = v->getXForFrame(i->getFrame()) - x;
         double distance = sqrt(yd*yd + xd*xd);
 
         if (nearestDistance == -1 || distance < nearestDistance) {
@@ -186,33 +201,34 @@ TextLayer::getFeatureDescription(LayerGeometryProvider *v, QPoint &pos) const
 {
     int x = pos.x();
 
-    if (!m_model || !m_model->getSampleRate()) return "";
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !model->getSampleRate()) return "";
 
-    TextModel::PointList points = getLocalPoints(v, x, pos.y());
+    EventVector points = getLocalPoints(v, x, pos.y());
 
     if (points.empty()) {
-        if (!m_model->isReady()) {
+        if (!model->isReady()) {
             return tr("In progress");
         } else {
             return "";
         }
     }
 
-    sv_frame_t useFrame = points.begin()->frame;
+    sv_frame_t useFrame = points.begin()->getFrame();
 
-    RealTime rt = RealTime::frame2RealTime(useFrame, m_model->getSampleRate());
+    RealTime rt = RealTime::frame2RealTime(useFrame, model->getSampleRate());
     
     QString text;
 
-    if (points.begin()->label == "") {
+    if (points.begin()->getLabel() == "") {
         text = QString(tr("Time:\t%1\nHeight:\t%2\nLabel:\t%3"))
             .arg(rt.toText(true).c_str())
-            .arg(points.begin()->height)
-            .arg(points.begin()->label);
+            .arg(points.begin()->getValue())
+            .arg(points.begin()->getLabel());
     }
 
     pos = QPoint(v->getXForFrame(useFrame),
-                 getYForHeight(v, points.begin()->height));
+                 getYForHeight(v, points.begin()->getValue()));
     return text;
 }
 
@@ -222,73 +238,40 @@ TextLayer::getFeatureDescription(LayerGeometryProvider *v, QPoint &pos) const
 bool
 TextLayer::snapToFeatureFrame(LayerGeometryProvider *v, sv_frame_t &frame,
                               int &resolution,
-                              SnapType snap) const
+                              SnapType snap, int ycoord) const
 {
-    if (!m_model) {
-        return Layer::snapToFeatureFrame(v, frame, resolution, snap);
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) {
+        return Layer::snapToFeatureFrame(v, frame, resolution, snap, ycoord);
     }
 
-    resolution = m_model->getResolution();
-    TextModel::PointList points;
+    // SnapLeft / SnapRight: return frame of nearest feature in that
+    // direction no matter how far away
+    //
+    // SnapNeighbouring: return frame of feature that would be used in
+    // an editing operation, i.e. closest feature in either direction
+    // but only if it is "close enough"
+
+    resolution = model->getResolution();
 
     if (snap == SnapNeighbouring) {
-        
-        points = getLocalPoints(v, v->getXForFrame(frame), -1);
+        EventVector points = getLocalPoints(v, v->getXForFrame(frame), -1);
         if (points.empty()) return false;
-        frame = points.begin()->frame;
+        frame = points.begin()->getFrame();
         return true;
     }    
 
-    points = m_model->getPoints(frame, frame);
-    sv_frame_t snapped = frame;
-    bool found = false;
-
-    for (TextModel::PointList::const_iterator i = points.begin();
-         i != points.end(); ++i) {
-
-        if (snap == SnapRight) {
-
-            if (i->frame > frame) {
-                snapped = i->frame;
-                found = true;
-                break;
-            }
-
-        } else if (snap == SnapLeft) {
-
-            if (i->frame <= frame) {
-                snapped = i->frame;
-                found = true; // don't break, as the next may be better
-            } else {
-                break;
-            }
-
-        } else { // nearest
-
-            TextModel::PointList::const_iterator j = i;
-            ++j;
-
-            if (j == points.end()) {
-
-                snapped = i->frame;
-                found = true;
-                break;
-
-            } else if (j->frame >= frame) {
-
-                if (j->frame - frame < frame - i->frame) {
-                    snapped = j->frame;
-                } else {
-                    snapped = i->frame;
-                }
-                found = true;
-                break;
-            }
-        }
+    Event e;
+    if (model->getNearestEventMatching
+        (frame,
+         [](Event) { return true; },
+         snap == SnapLeft ? EventSeries::Backward : EventSeries::Forward,
+         e)) {
+        frame = e.getFrame();
+        return true;
     }
 
-    frame = snapped;
-    return found;
+    return false;
 }
 
 int
@@ -308,18 +291,20 @@ TextLayer::getHeightForY(LayerGeometryProvider *v, int y) const
 void
 TextLayer::paint(LayerGeometryProvider *v, QPainter &paint, QRect rect) const
 {
-    if (!m_model || !m_model->isOK()) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !model->isOK()) return;
 
-    sv_samplerate_t sampleRate = m_model->getSampleRate();
+    sv_samplerate_t sampleRate = model->getSampleRate();
     if (!sampleRate) return;
 
 //    Profiler profiler("TextLayer::paint", true);
 
     int x0 = rect.left(), x1 = rect.right();
-    sv_frame_t frame0 = v->getFrameForX(x0);
-    sv_frame_t frame1 = v->getFrameForX(x1);
+    int overlap = ViewManager::scalePixelSize(150);
+    sv_frame_t frame0 = v->getFrameForX(x0 - overlap);
+    sv_frame_t frame1 = v->getFrameForX(x1 + overlap);
 
-    TextModel::PointList points(m_model->getPoints(frame0, frame1));
+    EventVector points(model->getEventsWithin(frame0, frame1 - frame0, 2));
     if (points.empty()) return;
 
     QColor brushColour(getBaseQColor());
@@ -332,10 +317,10 @@ TextLayer::paint(LayerGeometryProvider *v, QPainter &paint, QRect rect) const
     penColour = v->getForeground();
 
 //    SVDEBUG << "TextLayer::paint: resolution is "
-//              << m_model->getResolution() << " frames" << endl;
+//              << model->getResolution() << " frames" << endl;
 
     QPoint localPos;
-    TextModel::Point illuminatePoint(0);
+    Event illuminatePoint(0);
     bool shouldIlluminate = false;
 
     if (v->shouldIlluminateLocalFeatures(this, localPos)) {
@@ -349,18 +334,15 @@ TextLayer::paint(LayerGeometryProvider *v, QPainter &paint, QRect rect) const
     paint.save();
     paint.setClipRect(rect.x(), 0, rect.width() + boxMaxWidth, v->getPaintHeight());
     
-    for (TextModel::PointList::const_iterator i = points.begin();
+    for (EventVector::const_iterator i = points.begin();
          i != points.end(); ++i) {
 
-        const TextModel::Point &p(*i);
+        Event p(*i);
 
-        int x = v->getXForFrame(p.frame);
-        int y = getYForHeight(v, p.height);
+        int x = v->getXForFrame(p.getFrame());
+        int y = getYForHeight(v, p.getValue());
 
-        if (!shouldIlluminate ||
-            // "illuminatePoint != p"
-            TextModel::Point::Comparator()(illuminatePoint, p) ||
-            TextModel::Point::Comparator()(p, illuminatePoint)) {
+        if (!shouldIlluminate || illuminatePoint != p) {
             paint.setPen(penColour);
             paint.setBrush(brushColour);
         } else {
@@ -368,7 +350,7 @@ TextLayer::paint(LayerGeometryProvider *v, QPainter &paint, QRect rect) const
             paint.setPen(v->getBackground());
         }
 
-        QString label = p.label;
+        QString label = p.getLabel();
         if (label == "") {
             label = tr("<no text>");
         }
@@ -399,8 +381,8 @@ TextLayer::paint(LayerGeometryProvider *v, QPainter &paint, QRect rect) const
                        Qt::AlignLeft | Qt::AlignTop | Qt::TextWordWrap,
                        label);
 
-///        if (p.label != "") {
-///            paint.drawText(x + 5, y - paint.fontMetrics().height() + paint.fontMetrics().ascent(), p.label);
+///        if (p.getLabel() != "") {
+///            paint.drawText(x + 5, y - paint.fontMetrics().height() + paint.fontMetrics().ascent(), p.getLabel());
 ///        }
     }
 
@@ -415,23 +397,24 @@ TextLayer::drawStart(LayerGeometryProvider *v, QMouseEvent *e)
 {
 //    SVDEBUG << "TextLayer::drawStart(" << e->x() << "," << e->y() << ")" << endl;
 
-    if (!m_model) {
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) {
         SVDEBUG << "TextLayer::drawStart: no model" << endl;
         return;
     }
 
     sv_frame_t frame = v->getFrameForX(e->x());
     if (frame < 0) frame = 0;
-    frame = frame / m_model->getResolution() * m_model->getResolution();
+    frame = frame / model->getResolution() * model->getResolution();
 
     double height = getHeightForY(v, e->y());
 
-    m_editingPoint = TextModel::Point(frame, float(height), "");
+    m_editingPoint = Event(frame, float(height), "");
     m_originalPoint = m_editingPoint;
 
     if (m_editingCommand) finish(m_editingCommand);
-    m_editingCommand = new TextModel::EditCommand(m_model, "Add Label");
-    m_editingCommand->addPoint(m_editingPoint);
+    m_editingCommand = new ChangeEventsCommand(m_model.untyped, "Add Label");
+    m_editingCommand->add(m_editingPoint);
 
     m_editing = true;
 }
@@ -441,54 +424,58 @@ TextLayer::drawDrag(LayerGeometryProvider *v, QMouseEvent *e)
 {
 //    SVDEBUG << "TextLayer::drawDrag(" << e->x() << "," << e->y() << ")" << endl;
 
-    if (!m_model || !m_editing) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !m_editing) return;
 
     sv_frame_t frame = v->getFrameForX(e->x());
     if (frame < 0) frame = 0;
-    frame = frame / m_model->getResolution() * m_model->getResolution();
+    frame = frame / model->getResolution() * model->getResolution();
 
     double height = getHeightForY(v, e->y());
 
-    m_editingCommand->deletePoint(m_editingPoint);
-    m_editingPoint.frame = frame;
-    m_editingPoint.height = float(height);
-    m_editingCommand->addPoint(m_editingPoint);
+    m_editingCommand->remove(m_editingPoint);
+    m_editingPoint = m_editingPoint
+        .withFrame(frame)
+        .withValue(float(height));
+    m_editingCommand->add(m_editingPoint);
 }
 
 void
 TextLayer::drawEnd(LayerGeometryProvider *v, QMouseEvent *)
 {
 //    SVDEBUG << "TextLayer::drawEnd(" << e->x() << "," << e->y() << ")" << endl;
-    if (!m_model || !m_editing) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !m_editing) return;
 
     bool ok = false;
     QString label = QInputDialog::getText(v->getView(), tr("Enter label"),
                                           tr("Please enter a new label:"),
                                           QLineEdit::Normal, "", &ok);
 
+    m_editingCommand->remove(m_editingPoint);
+    
     if (ok) {
-        TextModel::RelabelCommand *command =
-            new TextModel::RelabelCommand(m_model, m_editingPoint, label);
-        m_editingCommand->addCommand(command);
-    } else {
-        m_editingCommand->deletePoint(m_editingPoint);
+        m_editingPoint = m_editingPoint
+            .withLabel(label);
+        m_editingCommand->add(m_editingPoint);
     }
 
     finish(m_editingCommand);
-    m_editingCommand = 0;
+    m_editingCommand = nullptr;
     m_editing = false;
 }
 
 void
 TextLayer::eraseStart(LayerGeometryProvider *v, QMouseEvent *e)
 {
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
     if (!getPointToDrag(v, e->x(), e->y(), m_editingPoint)) return;
 
     if (m_editingCommand) {
         finish(m_editingCommand);
-        m_editingCommand = 0;
+        m_editingCommand = nullptr;
     }
 
     m_editing = true;
@@ -502,21 +489,20 @@ TextLayer::eraseDrag(LayerGeometryProvider *, QMouseEvent *)
 void
 TextLayer::eraseEnd(LayerGeometryProvider *v, QMouseEvent *e)
 {
-    if (!m_model || !m_editing) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !m_editing) return;
 
     m_editing = false;
 
-    TextModel::Point p(0);
+    Event p;
     if (!getPointToDrag(v, e->x(), e->y(), p)) return;
-    if (p.frame != m_editingPoint.frame || p.height != m_editingPoint.height) return;
-
-    m_editingCommand = new TextModel::EditCommand
-        (m_model, tr("Erase Point"));
-
-    m_editingCommand->deletePoint(m_editingPoint);
+    if (p.getFrame() != m_editingPoint.getFrame() ||
+        p.getValue() != m_editingPoint.getValue()) return;
 
+    m_editingCommand = new ChangeEventsCommand(m_model.untyped, tr("Erase Point"));
+    m_editingCommand->remove(m_editingPoint);
     finish(m_editingCommand);
-    m_editingCommand = 0;
+    m_editingCommand = nullptr;
     m_editing = false;
 }
 
@@ -525,7 +511,8 @@ TextLayer::editStart(LayerGeometryProvider *v, QMouseEvent *e)
 {
 //    SVDEBUG << "TextLayer::editStart(" << e->x() << "," << e->y() << ")" << endl;
 
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
     if (!getPointToDrag(v, e->x(), e->y(), m_editingPoint)) {
         return;
@@ -536,7 +523,7 @@ TextLayer::editStart(LayerGeometryProvider *v, QMouseEvent *e)
 
     if (m_editingCommand) {
         finish(m_editingCommand);
-        m_editingCommand = 0;
+        m_editingCommand = nullptr;
     }
 
     m_editing = true;
@@ -545,42 +532,44 @@ TextLayer::editStart(LayerGeometryProvider *v, QMouseEvent *e)
 void
 TextLayer::editDrag(LayerGeometryProvider *v, QMouseEvent *e)
 {
-    if (!m_model || !m_editing) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !m_editing) return;
 
-    sv_frame_t frameDiff = v->getFrameForX(e->x()) - v->getFrameForX(m_editOrigin.x());
-    double heightDiff = getHeightForY(v, e->y()) - getHeightForY(v, m_editOrigin.y());
+    sv_frame_t frameDiff =
+        v->getFrameForX(e->x()) - v->getFrameForX(m_editOrigin.x());
+    double heightDiff =
+        getHeightForY(v, e->y()) - getHeightForY(v, m_editOrigin.y());
 
-    sv_frame_t frame = m_originalPoint.frame + frameDiff;
-    double height = m_originalPoint.height + heightDiff;
+    sv_frame_t frame = m_originalPoint.getFrame() + frameDiff;
+    double height = m_originalPoint.getValue() + heightDiff;
 
-//    sv_frame_t frame = v->getFrameForX(e->x());
     if (frame < 0) frame = 0;
-    frame = (frame / m_model->getResolution()) * m_model->getResolution();
-
-//    double height = getHeightForY(v, e->y());
+    frame = (frame / model->getResolution()) * model->getResolution();
 
     if (!m_editingCommand) {
-        m_editingCommand = new TextModel::EditCommand(m_model, tr("Drag Label"));
+        m_editingCommand = new ChangeEventsCommand(m_model.untyped, tr("Drag Label"));
     }
 
-    m_editingCommand->deletePoint(m_editingPoint);
-    m_editingPoint.frame = frame;
-    m_editingPoint.height = float(height);
-    m_editingCommand->addPoint(m_editingPoint);
+    m_editingCommand->remove(m_editingPoint);
+    m_editingPoint = m_editingPoint
+        .withFrame(frame)
+        .withValue(float(height));
+    m_editingCommand->add(m_editingPoint);
 }
 
 void
 TextLayer::editEnd(LayerGeometryProvider *, QMouseEvent *)
 {
 //    SVDEBUG << "TextLayer::editEnd(" << e->x() << "," << e->y() << ")" << endl;
-    if (!m_model || !m_editing) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model || !m_editing) return;
 
     if (m_editingCommand) {
 
         QString newName = m_editingCommand->getName();
 
-        if (m_editingPoint.frame != m_originalPoint.frame) {
-            if (m_editingPoint.height != m_originalPoint.height) {
+        if (m_editingPoint.getFrame() != m_originalPoint.getFrame()) {
+            if (m_editingPoint.getValue() != m_originalPoint.getValue()) {
                 newName = tr("Move Label");
             } else {
                 newName = tr("Move Label Horizontally");
@@ -593,28 +582,31 @@ TextLayer::editEnd(LayerGeometryProvider *, QMouseEvent *)
         finish(m_editingCommand);
     }
     
-    m_editingCommand = 0;
+    m_editingCommand = nullptr;
     m_editing = false;
 }
 
 bool
 TextLayer::editOpen(LayerGeometryProvider *v, QMouseEvent *e)
 {
-    if (!m_model) return false;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return false;
 
-    TextModel::Point text(0);
+    Event text;
     if (!getPointToDrag(v, e->x(), e->y(), text)) return false;
 
-    QString label = text.label;
+    QString label = text.getLabel();
 
     bool ok = false;
     label = QInputDialog::getText(v->getView(), tr("Enter label"),
                                   tr("Please enter a new label:"),
                                   QLineEdit::Normal, label, &ok);
-    if (ok && label != text.label) {
-        TextModel::RelabelCommand *command =
-            new TextModel::RelabelCommand(m_model, text, label);
-        CommandHistory::getInstance()->addCommand(command);
+    if (ok && label != text.getLabel()) {
+        ChangeEventsCommand *command =
+            new ChangeEventsCommand(m_model.untyped, tr("Re-Label Point"));
+        command->remove(text);
+        command->add(text.withLabel(label));
+        finish(command);
     }
 
     return true;
@@ -623,23 +615,20 @@ TextLayer::editOpen(LayerGeometryProvider *v, QMouseEvent *e)
 void
 TextLayer::moveSelection(Selection s, sv_frame_t newStartFrame)
 {
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
-    TextModel::EditCommand *command =
-        new TextModel::EditCommand(m_model, tr("Drag Selection"));
+    ChangeEventsCommand *command =
+        new ChangeEventsCommand(m_model.untyped, tr("Drag Selection"));
 
-    TextModel::PointList points =
-        m_model->getPoints(s.getStartFrame(), s.getEndFrame());
+    EventVector points =
+        model->getEventsStartingWithin(s.getStartFrame(), s.getDuration());
 
-    for (TextModel::PointList::iterator i = points.begin();
-         i != points.end(); ++i) {
-
-        if (s.contains(i->frame)) {
-            TextModel::Point newPoint(*i);
-            newPoint.frame = i->frame + newStartFrame - s.getStartFrame();
-            command->deletePoint(*i);
-            command->addPoint(newPoint);
-        }
+    for (Event p: points) {
+        command->remove(p);
+        Event moved = p.withFrame(p.getFrame() +
+                                  newStartFrame - s.getStartFrame());
+        command->add(moved);
     }
 
     finish(command);
@@ -648,32 +637,27 @@ TextLayer::moveSelection(Selection s, sv_frame_t newStartFrame)
 void
 TextLayer::resizeSelection(Selection s, Selection newSize)
 {
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
-    TextModel::EditCommand *command =
-        new TextModel::EditCommand(m_model, tr("Resize Selection"));
+    ChangeEventsCommand *command =
+        new ChangeEventsCommand(m_model.untyped, tr("Resize Selection"));
 
-    TextModel::PointList points =
-        m_model->getPoints(s.getStartFrame(), s.getEndFrame());
+    EventVector points =
+        model->getEventsStartingWithin(s.getStartFrame(), s.getDuration());
 
-    double ratio =
-        double(newSize.getEndFrame() - newSize.getStartFrame()) /
-        double(s.getEndFrame() - s.getStartFrame());
+    double ratio = double(newSize.getDuration()) / double(s.getDuration());
+    double oldStart = double(s.getStartFrame());
+    double newStart = double(newSize.getStartFrame());
+    
+    for (Event p: points) {
 
-    for (TextModel::PointList::iterator i = points.begin();
-         i != points.end(); ++i) {
+        double newFrame = (double(p.getFrame()) - oldStart) * ratio + newStart;
 
-        if (s.contains(i->frame)) {
-
-            double target = double(i->frame);
-            target = double(newSize.getStartFrame()) + 
-                target - double(s.getStartFrame()) * ratio;
-
-            TextModel::Point newPoint(*i);
-            newPoint.frame = lrint(target);
-            command->deletePoint(*i);
-            command->addPoint(newPoint);
-        }
+        Event newPoint = p
+            .withFrame(lrint(newFrame));
+        command->remove(p);
+        command->add(newPoint);
     }
 
     finish(command);
@@ -682,17 +666,17 @@ TextLayer::resizeSelection(Selection s, Selection newSize)
 void
 TextLayer::deleteSelection(Selection s)
 {
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
-    TextModel::EditCommand *command =
-        new TextModel::EditCommand(m_model, tr("Delete Selection"));
+    ChangeEventsCommand *command =
+        new ChangeEventsCommand(m_model.untyped, tr("Delete Selection"));
 
-    TextModel::PointList points =
-        m_model->getPoints(s.getStartFrame(), s.getEndFrame());
+    EventVector points =
+        model->getEventsStartingWithin(s.getStartFrame(), s.getDuration());
 
-    for (TextModel::PointList::iterator i = points.begin();
-         i != points.end(); ++i) {
-        if (s.contains(i->frame)) command->deletePoint(*i);
+    for (Event p: points) {
+        command->remove(p);
     }
 
     finish(command);
@@ -701,27 +685,25 @@ TextLayer::deleteSelection(Selection s)
 void
 TextLayer::copy(LayerGeometryProvider *v, Selection s, Clipboard &to)
 {
-    if (!m_model) return;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return;
 
-    TextModel::PointList points =
-        m_model->getPoints(s.getStartFrame(), s.getEndFrame());
+    EventVector points =
+        model->getEventsStartingWithin(s.getStartFrame(), s.getDuration());
 
-    for (TextModel::PointList::iterator i = points.begin();
-         i != points.end(); ++i) {
-        if (s.contains(i->frame)) {
-            Clipboard::Point point(i->frame, i->height, i->label);
-            point.setReferenceFrame(alignToReference(v, i->frame));
-            to.addPoint(point);
-        }
+    for (Event p: points) {
+        to.addPoint(p.withReferenceFrame(alignToReference(v, p.getFrame())));
     }
 }
 
 bool
-TextLayer::paste(LayerGeometryProvider *v, const Clipboard &from, sv_frame_t /* frameOffset */, bool /* interactive */)
+TextLayer::paste(LayerGeometryProvider *v, const Clipboard &from,
+                 sv_frame_t /* frameOffset */, bool /* interactive */)
 {
-    if (!m_model) return false;
+    auto model = ModelById::getAs<TextModel>(m_model);
+    if (!model) return false;
 
-    const Clipboard::PointList &points = from.getPoints();
+    const EventVector &points = from.getPoints();
 
     bool realign = false;
 
@@ -742,23 +724,22 @@ TextLayer::paste(LayerGeometryProvider *v, const Clipboard &from, sv_frame_t /*
         }
     }
 
-    TextModel::EditCommand *command =
-        new TextModel::EditCommand(m_model, tr("Paste"));
+    ChangeEventsCommand *command =
+        new ChangeEventsCommand(m_model.untyped, tr("Paste"));
 
     double valueMin = 0.0, valueMax = 1.0;
-    for (Clipboard::PointList::const_iterator i = points.begin();
+    for (EventVector::const_iterator i = points.begin();
          i != points.end(); ++i) {
-        if (i->haveValue()) {
+        if (i->hasValue()) {
             if (i->getValue() < valueMin) valueMin = i->getValue();
             if (i->getValue() > valueMax) valueMax = i->getValue();
         }
     }
     if (valueMax < valueMin + 1.0) valueMax = valueMin + 1.0;
 
-    for (Clipboard::PointList::const_iterator i = points.begin();
+    for (EventVector::const_iterator i = points.begin();
          i != points.end(); ++i) {
         
-        if (!i->haveFrame()) continue;
         sv_frame_t frame = 0;
         
         if (!realign) {
@@ -767,7 +748,7 @@ TextLayer::paste(LayerGeometryProvider *v, const Clipboard &from, sv_frame_t /*
 
         } else {
 
-            if (i->haveReferenceFrame()) {
+            if (i->hasReferenceFrame()) {
                 frame = i->getReferenceFrame();
                 frame = alignFromReference(v, frame);
             } else {
@@ -775,23 +756,25 @@ TextLayer::paste(LayerGeometryProvider *v, const Clipboard &from, sv_frame_t /*
             }
         }
 
-        TextModel::Point newPoint(frame);
+        Event p = i->withFrame(frame);
 
-        if (i->haveValue()) {
-            newPoint.height = float((i->getValue() - valueMin) / (valueMax - valueMin));
+        Event newPoint = p;
+        if (p.hasValue()) {
+            newPoint = newPoint.withValue(float((i->getValue() - valueMin) /
+                                                (valueMax - valueMin)));
         } else {
-            newPoint.height = 0.5f;
+            newPoint = newPoint.withValue(0.5f);
         }
 
-        if (i->haveLabel()) {
-            newPoint.label = i->getLabel();
-        } else if (i->haveValue()) {
-            newPoint.label = QString("%1").arg(i->getValue());
-        } else {
-            newPoint.label = tr("New Point");
+        if (!p.hasLabel()) {
+            if (p.hasValue()) {
+                newPoint = newPoint.withLabel(QString("%1").arg(p.getValue()));
+            } else {
+                newPoint = newPoint.withLabel(tr("New Point"));
+            }
         }
         
-        command->addPoint(newPoint);
+        command->add(newPoint);
     }
 
     finish(command);
